/* grid.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#ifndef SEEK_SET
#  define SEEK_SET 0
#endif
#include "globals.h"
#include "memory.h"




/* Format of VIS-5D Compressed Grid Files:

   Information                          Size (in bytes) , Type
 --------------------------------------------------------------------------
   ID                                   4   int
      if (ID==0x80808080)
         TIMES = 300
         PARMS = 20
      if (ID==0x80808081)
         TIMES = 400
	 PARMS = 30

   Number of time steps (NumTimes)      4   int
   Number of parameters (NumParms)      4   int
   Number of rows (Nr)                  4   int
   Number of columns (Nc)               4   int
   Number of levels (Nl)                4   int
   North latitude (LatN)                4   float
   West longitude (LonW)                4   float
   Top height (HgtT)                    4   float
   Latitude inc/grid pt. (LatInc)       4   float
   Longitude inc/grid pt. (LonInc)      4   float
   Height inc/grid pt. (HgtInc)         4   float

   Day values for each time step        4 * TIMES   int
   Time values for each time step       4 * TIMES   int
   Paramenter Names                     4 * PARMS   char

   FOR i=1 TO NumTimes DO

      FOR j=1 TO NumParms DO

         GA value                       sizeof(float)

         GB value                       sizeof(float)

         Compressed Grid Data           ((Nr*Nc*Nl+3)/4)*4  (multiple of 4)

      ENDFOR

   ENDFOR

 --------------------------------------------------------------------------
*/



/*** module variables: ***/

static float Ga[MAXTIMES][MAXPARMS];
static float Gb[MAXTIMES][MAXPARMS];
static int_1 *Grid[MAXTIMES][MAXPARMS];




/*** load_dataset *****************************************************
   Load the dataset.
   Input:  gridfile - filename of the compressed grid file to load.
   Return:  1 = success, 0 = error.
**********************************************************************/
int load_dataset( gridfile )
char *gridfile;
{
   FILE *f;
   int i, j, id, pos, ip,it;
   int first;
   float ga, gb, min, max, latn, lonw, hgtt;
   float dlat, dlon;
   int gridsize;
   int gridtimes, gridparms;

   /* open grid file */
   f = fopen( gridfile, "r" );
   if (!f) {
      /* try uppercase name */
      for (i=strlen(gridfile)-1;i>=0;i--) {
	 if (gridfile[i]=='/')
	    break;
	 else if (islower(gridfile[i]))
	    gridfile[i] += 'A'-'a';
      }
      f = fopen( gridfile, "r" );
      if (!f) {
	 printf("File not found\n");
	 return 0;
      }
   }

   /* read header info */
   read_int4( f, &id );
   if (id==0x80808080) {
      /* old comp file: 20 vars, 300 times */
      gridtimes = 300;
      gridparms = 20;
   }
   else if (id==0x80808081) {
      /* new comp file: 30 vars, 400 times */
      gridtimes = 400;
      gridparms = 30;
   }
   else {
      printf("Not a VIS-5D file\n");
      fclose(f);
      return 0;
   }

   assert(gridtimes<=MAXTIMES && gridparms<=MAXPARMS);

   read_int4( f, &NumTimes );
   if (NumTimes<1) {
      printf("Error:  Number of time steps = %d\n", NumTimes );
      fclose(f);
      return 0;
   }
   read_int4( f, &NumParms );
   if (NumParms<1) {
      printf("Error:  Number of physical variables = %d\n", NumParms );
      fclose(f);
      return 0;
   }
   read_int4( f, &Nr );
   read_int4( f, &Nc );
   read_int4( f, &Nl );
   if (Nr<2 || Nc<2 || Nl<2) {
      printf("Bad grid dimension(s):  rows=%d columns=%d levels=%d\n", Nr,Nc,Nl);
      fclose(f);
      return 0;
   }
   read_float4( f, &latn );
   read_float4( f, &lonw );
   read_float4( f, &hgtt );
   read_float4( f, &LatInc );
   read_float4( f, &LonInc );
   read_float4( f, &HgtInc );
   if (LatInc<=0.0 || LonInc<=0.0 || HgtInc<=0.0) {
      printf("Bad grid increment(s):  LatInc=%f LonInc=%f HgtInc=%f\n",
	     LatInc, LonInc, HgtInc );
      fclose(f);
      return 0;
   }

   for (i=0;i<gridtimes;i++) {
      read_int4( f, &DayStamp[i] );
   }
   for (i=0;i<gridtimes;i++) {
      read_int4( f, &TimeStamp[i] );
   }
   /* read parm names */
   for (i=0;i<gridparms;i++) {
      if ( !fread( ParmName[i], 4, 1, f) ) {
	 printf("Error reading parameter names.\n");
	 fclose(f);
	 return NULL;
      }
      /* remove trailing spaces */
      for (j=3;j>0;j--) {
	 if (ParmName[i][j]==' ') {
	    ParmName[i][j] = 0;
	 }
	 else {
	    break;
	 }
      }
   }

   for (ip=0;ip<NumParms;ip++) {
      MinVal[ip] = 999999.9;
      MaxVal[ip] = -999999.9;
   }

   gridsize = ((Nr*Nc*Nl+3)/4)*4;

   /* read the grids and their ga,gb values, find min and max values */
   for (it=0;it<NumTimes;it++) {
      for (ip=0;ip<NumParms;ip++) {
	 read_float4( f, &ga );
	 read_float4( f, &gb );
	 Ga[it][ip] = ga;
	 Gb[it][ip] = gb;
	 Grid[it][ip] = (int_1 *) pallocate( gridsize );
	 if (!Grid[it][ip]) {
	    printf("Out of memory while reading grid file.\n");
	    fclose(f);
	    return 0;
	 }
	 if (fread( Grid[it][ip], sizeof(int_1), gridsize, f )!=gridsize) {
	    printf("ERROR while reading grid file, may be corrupted.\n");
	    return 0;
	 }
	 min = -(125.0+gb)/ga;
	 max = (125.0-gb)/ga;
	 if (min<MinVal[ip])  MinVal[ip] = min;
	 if (max>MaxVal[ip])  MaxVal[ip] = max;
      }
   }

   fclose(f);

   /* Derive other global values */

   /* calculate elapsed time for each time since initial time */
   first = DayStamp[0]* 24*60*60 + TimeStamp[0];
   for (it=0;it<NumTimes;it++) {
      Elapsed[it] = DayStamp[it]*24*60*60+TimeStamp[it] - first;
   }

   /* find U, V, W variables */
   Uparm = Vparm = Wparm = -1;
   for (ip=0;ip<NumParms;ip++) {
      if (ParmName[ip][0]=='U')
	 Uparm = ip;
      if (ParmName[ip][0]=='V')
	 Vparm = ip;
      if (ParmName[ip][0]=='W')
	 Wparm = ip;
   }
   if (Uparm>=0 && Vparm>=0 && Wparm>=0)
      WindAvailable = 1;
   else
      WindAvailable = 0;

   /* simulated moving frame of reference */
#ifdef LEAVEOUT
   if (filename[0]=='V') {
      dlat = 0.004697 * 2.0;
      dlon = 0.03096 * 2.0;
   }
   else {
      dlat = -LatInc*(Nr-1) / (float) NumTimes / 2.0;
      dlon = -LonInc*(Nc-1) / (float) NumTimes / 2.0;
   }
#endif
   dlat = dlon = 0.0;

   for (i=0;i<NumTimes;i++) {
      NorthLat[i] = latn;
      WestLon[i] = lonw;
      TopHgt[i] = hgtt;

      latn += dlat;
      lonw += dlon;
   }

   /* calculate south, east, and bottom bounds for each timestep */
   for (it=0;it<NumTimes;it++) {
      SouthLat[it] = NorthLat[it]-(Nr-1)*LatInc;
      EastLon[it] = WestLon[it]-(Nc-1)*LonInc;
      BottomHgt[it] = TopHgt[it]-(Nl-1)*HgtInc;
   }

   /* calculate total domain size */
   NorthBound = NorthLat[0];
   SouthBound = SouthLat[0];
   WestBound = WestLon[0];
   EastBound = EastLon[0];
   TopBound = TopHgt[0];
   BottomBound = BottomHgt[0];
   StaticDomain = 1;
   for (it=1;it<NumTimes;it++) {
      if (NorthLat[it]>NorthBound)
	 NorthBound = NorthLat[it];
      if (SouthLat[it]>SouthBound)
	 SouthBound = SouthLat[it];
      if (WestLon[it]>WestBound)
	 WestBound = WestLon[it];
      if (EastLon[it]>EastBound)
	 EastBound = EastLon[it];
      if (WestLon[it]!=WestLon[0] || EastLon[it]!=EastLon[0]
	  || NorthLat[it]!=NorthLat[0] || SouthLat[it]!=SouthLat[0]) {
	 StaticDomain = 0;
      }
   }
   return 1;
}




/*** get_grid *********************************************************
   Return a pointer to the uncompressed data for a 3-D grid.
   Input:  it, ip - time and parameter of grid wanted.
   Return:  pointer to float data
**********************************************************************/
float *get_grid( it, ip )
int it, ip;
{
   float *data;

   assert( it>=0 && it<NumTimes );
   assert( ip>=0 && ip<NumParms );

   data = (float *) allocate( Nr*Nc*Nl*sizeof(float) );
   assert(data!=NULL);

   decompress( data, Grid[it][ip], Nr*Nc*Nl, Ga[it][ip], Gb[it][ip] );
   return data;
}



/*** get_compressed_grid **********************************************
   Return a pointer to the compressed data for a 3-D grid.
   Input: it, ip - time and parameter of grid wanted.
          ga, gb - pointer to floats.
   Output: ga, gb - values to use for decompressing.
   Return:  pointer to int_1 data
**********************************************************************/
int_1 *get_compressed_grid( it, ip, ga, gb )
int it, ip;
float *ga, *gb;
{
   *ga = Ga[it][ip];
   *gb = Gb[it][ip];
   return Grid[it][ip];
}




/*** release_grid *****************************************************
   Release an uncompressed grid.
   Input:  time, parm - the timestep and parameter of the grid.
           data - the grid data returned by get_grid().
**********************************************************************/
void release_grid( it, ip, data )
int it, ip;
float *data;
{
   assert( it>=0 && it<NumTimes );
   assert( ip>=0 && ip<NumParms );

   deallocate( data, Nr*Nc*Nl*sizeof(float) );
}



/*** release_compressed_grid ******************************************
   Release a compressed grid.
   Input:  it, ip - the timestep and parameter to grid to release.
**********************************************************************/
void release_compressed_grid( it, ip )
int it, ip;
{
   /* nothing */
}




/*** decompress *****************************************************
   Convert an array of chars to floats.  This function is only
   called from get_grid() and vectorized for the stellar.
********************************************************************/
static int decompress( data, compdata, num, ga, gb )
float data[];
int_1 compdata[];
int num;
float ga, gb;
{
   int i;
   /*$dir no_recurrence */  /* stellar:  fully vectorized */
   for (i=0;i<num;i++) {
      if (compdata[i]==127)
	data[i] = 1.0e35;
      else
	data[i] = ( (float) (int) compdata[i] - gb ) / ga;
   }
}




/*** read_int4 ********************************************************
   Read a 4-byte integer from a file.
   Input:  f - the file to read from
           i - pointer to integer to put result into.
   Returned:  0 : error
             !0 : success
**********************************************************************/
static int read_int4( f, i )
FILE *f;
int *i;
{
   if ( fread( i, 4, 1, f ) ) {
      return 1;
   }
   return 0;
}



/*** read_float4 ******************************************************
   Read a 4-byte IEEE float from a file.
   Input:  f - the file to read from.
           x - pointer to float to put result into.
   Returned:  0 : error
             !0 : success
**********************************************************************/
static int read_float4( f, x )
FILE *f;
float *x;
{
   if ( fread( x, 4, 1, f ) ) {
      return 1;
    }
    return 0;
}


